from typing import Optional, Callable

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import PolynomialFeatures, FunctionTransformer
from sklearn.utils.validation import check_is_fitted

from ..base import FeatureMixin, add_class_name

__all__ = [
    "Shift",
    "MovingAverage",
    "MovingCustomFunction",
    "Polynomial",
    "Exogenous",
    "CustomFeature",
]


class Shift(BaseEstimator, TransformerMixin, FeatureMixin):
    """Perform a shift of a DataFrame of size equal to ``shift``.

    Parameters
    ----------
    shift : int, optional, default: ``1``
        How much to shift.

    Notes
    -----
    The ``shift`` parameter can also accept negative values. However, this should be
    used carefully, since if the resulting feature is used for training or testing it
    might generate a leak from the feature.

    Examples
    --------
    >>> import pandas as pd
    >>> from gtime.feature_extraction import Shift
    >>> ts = pd.DataFrame([0, 1, 2, 3, 4, 5])
    >>> shift = Shift(shift=3)
    >>> shift.fit_transform(ts)
       0__Shift
    0       NaN
    1       NaN
    2       NaN
    3       0.0
    4       1.0
    5       2.0

    """

    def __init__(self, shift: int = 1):
        super().__init__()
        self.shift = shift

    def fit(self, X, y=None):
        """Fit the estimator.

        Parameters
        ----------
        X : pd.DataFrame, shape (n_samples, n_features)
            Input data.

        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Returns
        -------
        self : object
            Returns self.
        """
        self.columns_ = X.columns.values
        return self

    @add_class_name
    def transform(self, time_series: pd.DataFrame) -> pd.DataFrame:
        """Create a shifted version of ``time_series``.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, 1), required
            The DataFrame to shift.

        Returns
        -------
        time_series_t : pd.DataFrame, shape (n_samples, 1)
            The shifted version of the original ``time_series``.

        """
        check_is_fitted(self)

        time_series_shifted = time_series.shift(self.shift)
        return time_series_shifted


class MovingAverage(BaseEstimator, TransformerMixin, FeatureMixin):
    """For each row in ``time_series``, compute the moving average of the previous
    ``window_size`` rows. If there are not enough rows, the value is ``Nan``.

    Parameters
    ----------
    window_size : int, optional, default: ``1``
        The number of previous points on which to compute the moving average.

    Examples
    --------
    >>> import pandas as pd
    >>> from gtime.feature_extraction import MovingAverage
    >>> ts = pd.DataFrame([0, 1, 2, 3, 4, 5])
    >>> mv_avg = MovingAverage(window_size=2)
    >>> mv_avg.fit_transform(ts)
       0__MovingAverage
    0               NaN
    1               0.5
    2               1.5
    3               2.5
    4               3.5
    5               4.5

    """

    def __init__(self, window_size: int = 1):
        super().__init__()
        self.window_size = window_size

    def fit(self, X, y=None):
        """Fit the estimator.

        Parameters
        ----------
        X : pd.DataFrame, shape (n_samples, n_features)
            Input data.

        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Returns
        -------
        self : object
            Returns self.

        """
        self.columns_ = X.columns.values
        return self

    @add_class_name
    def transform(self, time_series: pd.DataFrame) -> pd.DataFrame:
        """Compute the moving average, for every row of ``time_series``, of the previous
        ``window_size`` elements.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, 1), required
            The DataFrame on which to compute the rolling moving average

        Returns
        -------
        time_series_t : pd.DataFrame, shape (n_samples, 1)
            A DataFrame, with the same length as ``time_series``, containing the rolling
            moving average for each element.

        """
        check_is_fitted(self)

        time_series_mvg_avg = time_series.rolling(self.window_size).mean()
        return time_series_mvg_avg


class MovingCustomFunction(BaseEstimator, TransformerMixin, FeatureMixin):
    """For each row in ``time_series``, compute the moving custom function of the
    previous ``window_size`` rows. If there are not enough rows, the value is ``Nan``.

    Parameters
    ----------
    custom_feature_function : Callable, required.
        The function to use to generate a ``pd.DataFrame`` containing the feature.

    window_size : int, optional, default: ``1``
        The number of previous points on which to compute the custom function.

    raw : bool, optional, default: ``True``
        - ``False`` : passes each row or column as a Series to the function.
        - ``True`` or ``None`` : the passed function will receive ndarray objects
        instead.
         If you are just applying a NumPy reduction function this will achieve much
         better performance. Credits: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.core.window.Rolling.apply.html

    Examples
    --------
    >>> import pandas as pd
    >>> import numpy as np
    >>> from gtime.feature_extraction import MovingCustomFunction
    >>> ts = pd.DataFrame([0, 1, 2, 3, 4, 5])
    >>> mv_custom = MovingCustomFunction(np.max, window_size=2)
    >>> mv_custom.fit_transform(ts)
       0__MovingCustomFunction
    0                      NaN
    1                      1.0
    2                      2.0
    3                      3.0
    4                      4.0
    5                      5.0

    """

    def __init__(
        self, custom_feature_function: Callable, window_size: int = 1, raw: bool = True,
    ):
        super().__init__()
        self.custom_feature_function = custom_feature_function
        self.window_size = window_size
        self.raw = raw

    def fit(self, time_series: pd.DataFrame, y=None):
        """Fit the estimator.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, n_features)
            Input data.

        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Returns
        -------
        self : object
            Returns self.

        """
        self.columns_ = time_series.columns.values
        return self

    @add_class_name
    def transform(self, time_series: pd.DataFrame) -> pd.DataFrame:
        """For every row of ``time_series``, compute the moving custom function of the
         previous ``window_size`` elements.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, 1), required
            The DataFrame on which to compute the rolling moving custom function.

        Returns
        -------
        time_series_t : pd.DataFrame, shape (n_samples, 1)
            A DataFrame, with the same length as ``time_series``, containing the rolling
            moving custom function for each element.

        """
        check_is_fitted(self)

        time_series_mvg_cust = time_series.rolling(self.window_size).apply(
            self.custom_feature_function, raw=self.raw
        )
        time_series_t = time_series_mvg_cust
        return time_series_t


class Polynomial(PolynomialFeatures, FeatureMixin):
    """Compute the polynomial feature_extraction, of a degree equal to the input
    ``degree``.
    Wrapper of ``sklearn.preprocessing.PolynomialFeatures`` but returns a
    ``pd.DataFrame``.

    Parameters
    ----------
    degree : int, optional, default: ``2``
        The degree of the polynomial feature_extraction.

    Examples
    --------
    >>> import pandas as pd
    >>> from gtime.feature_extraction import Polynomial
    >>> ts = pd.DataFrame([0, 1, 2, 3, 4, 5])
    >>> pol = Polynomial(degree=3)
    >>> pol.fit_transform(ts)
       0__Polynomial  1__Polynomial  2__Polynomial  3__Polynomial
    0            1.0            0.0            0.0            0.0
    1            1.0            1.0            1.0            1.0
    2            1.0            2.0            4.0            8.0
    3            1.0            3.0            9.0           27.0
    4            1.0            4.0           16.0           64.0
    5            1.0            5.0           25.0          125.0

    """

    def __init__(self, degree: int = 2):
        super().__init__()
        self.degree = degree

    def fit(self, time_series: pd.DataFrame, y=None):
        """Fit the estimator.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, n_features)
            Input data.

        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Returns
        -------
        self : object
            Returns self.

        """
        self.columns_ = time_series.columns.values
        return super().fit(time_series)

    @add_class_name
    def transform(self, time_series: pd.DataFrame) -> pd.DataFrame:
        """Compute the polynomial feature_extraction of ``time_series``, up to a degree
        equal to ``degree``.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, 1), required
            The input DataFrame. Used only for its index.

        Returns
        -------
        time_series_t : pd.DataFrame, shape (n_samples, 1)
            The computed polynomial feature_extraction.

        """
        check_is_fitted(self)

        X_t = super().transform(time_series)
        X_t_df = pd.DataFrame(
            data=X_t, columns=self.get_feature_names(), index=time_series.index
        )

        return X_t_df


class Exogenous(BaseEstimator, TransformerMixin, FeatureMixin):
    """Reindex ``exogenous_time_series`` with the index of ``time_series``. To check the
    documentation of ``pandas.DataFrame.reindex`` and to see which type of
    ``method`` are available, please refer to the pandas `documentation
    <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.reindex.html>`_.

    Parameters
    ----------
    exogenous_time_series : pd.DataFrame, shape (n_samples, 1), required
        The time series to reindex

    method : str, optional, default: ``None``
        The method used to re-index. This must be a method used by the
        ``pandas.DataFrame.reindex`` method.

    Examples
    --------
    >>> import pandas as pd
    >>> from gtime.feature_extraction import Exogenous
    >>> ts = pd.DataFrame([0, 1, 2, 3, 4, 5], index=[3, 4, 5, 6, 7, 8])
    >>> exog_ts = pd.DataFrame([10, 8, 1, 3, 2, 7])
    >>> exog = Exogenous(exog_ts)
    >>> exog.fit_transform(ts)
       0__Exogenous
    3           3.0
    4           2.0
    5           7.0
    6           NaN
    7           NaN
    8           NaN

    >>> exog = Exogenous(exog_ts, method="nearest")
    >>> exog.fit_transform(ts)
       0__Exogenous
    3             3
    4             2
    5             7
    6             7
    7             7
    8             7

    """

    def __init__(
        self, exogenous_time_series: pd.DataFrame, method: Optional[str] = None,
    ):
        super().__init__()
        self.method = method
        self.exogenous_time_series = exogenous_time_series

    def fit(self, time_series, y=None):
        """Fit the estimator.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, n_features)
            Input data.

        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Returns
        -------
        self : object
            Returns self.

        """
        self.columns_ = time_series.columns.values
        return self

    @add_class_name
    def transform(self, time_series: pd.DataFrame) -> pd.DataFrame:
        """Reindex the ``exogenous_time_series`` with the index of ``time_series``.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, 1), required
            The input DataFrame. Used only for its index.

        Returns
        -------
        time_series_t :  pd.DataFrame, shape (n_samples, 1)
            The original ``exogenous_time_series``, re-indexed with the index
            of ``time_series``.

        """
        check_is_fitted(self)

        exog_feature = self.exogenous_time_series.reindex(
            index=time_series.index, method=self.method
        )

        return exog_feature


class CustomFeature(FunctionTransformer, FeatureMixin):
    """Constructs a transformer from an arbitrary callable.
    This transformer is a wrapper of ``sklearn.preprocessing.FunctionTransformer`` but returns a ``pd.Dataframe``.

    Parameters
    ----------
    func : Callable, required.
        The function to use to generate a ``pd.DataFrame`` containing the feature.

    kwargs : ``object``, optional.
        Optional arguments to pass to the transform method.

    Examples
    --------
    >>> import pandas as pd
    >>> from gtime.feature_extraction import CustomFeature
    >>> def custom_function(X, power):
    ...     return X**power
    >>> X = pd.DataFrame([0, 1, 2, 3, 4, 5])
    >>> custom_feature = CustomFeature(custom_function, power=3)
    >>> custom_feature.fit_transform(X)
       0__CustomFeature
    0                 0
    1                 1
    2                 8
    3                27
    4                64
    5               125

    """

    def __init__(
        self, func: Callable, **kwargs: object,
    ):
        super().__init__(inverse_func=None, func=func, kw_args=kwargs)

    def fit(self, time_series: pd.DataFrame, y=None) -> "CustomFeature":
        """Fit the estimator.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, n_features)
            Input data.

        y : None
            There is no need of a target in a transformer, yet the pipeline API requires
            this parameter.

        Returns
        -------
        self : object
            Returns self.

        """
        self.columns_ = time_series.columns.values
        return super().fit(time_series)

    @add_class_name
    def transform(self, time_series: Optional[pd.DataFrame] = None) -> pd.DataFrame:
        """Generate a ``pd.DataFrame``, given ``time_series`` as input to the
        ``func``, as well as other optional arguments.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, 1), optional, default: ``None``
            The DataFrame on which to apply the the custom function.

        Returns
        -------
        X_t_df : pd.DataFrame, shape (length, 1)
            A DataFrame containing the generated feature.

        """
        check_is_fitted(self)
        return super().transform(time_series)



from typing import Union, List

import numpy as np
import pandas as pd
from scipy.stats import gmean


def _check_input(y_true: np.ndarray, y_pred: np.ndarray) -> None:
    if len(y_pred) != len(y_true):
        raise ValueError(
            f"The arrays must have the same length, but they "
            f"have length {len(y_pred)} and {len(y_true)}."
        )
    if np.isnan(y_true).any() or np.isnan(y_pred).any():
        raise ValueError(
            "The two arrays should not contain Nan values, but they are "
            f"{y_true}, {y_pred}."
        )

    if np.isinf(y_true).any() or np.isinf(y_pred).any():
        raise ValueError(
            "The two arrays should not contain Inf values, but they are "
            f"{y_true}, {y_pred}."
        )


def _convert_to_ndarray(
    y_true: Union[pd.DataFrame, List, np.ndarray],
    y_pred: Union[pd.DataFrame, List, np.ndarray],
):
    if isinstance(y_true, pd.DataFrame):
        y_true = y_true.values
    elif isinstance(y_true, List):
        y_true = np.array(y_true)

    if isinstance(y_pred, pd.DataFrame):
        y_pred = y_pred.values
    elif isinstance(y_pred, List):
        y_pred = np.array(y_pred)

    return y_true, y_pred


def smape(
    y_true: Union[pd.DataFrame, List, np.ndarray],
    y_pred: Union[pd.DataFrame, List, np.ndarray],
) -> float:
    """Compute the 'Symmetric Mean Absolute Percentage Error' (SMAPE) between two
    vectors. Documentation
    `here <https://en.wikipedia.org/wiki/Symmetric_mean_absolute_percentage_error>_`.

    Parameters
    ----------
    y_true : array-like, shape (length, 1), required
        The first vector.

    y_pred : array-like, shape (length, 1), required
        The second vector.

    Returns
    -------
    smape : float
        The smape between the two input vectors.

    Examples
    --------
    >>> from gtime.metrics import smape
    >>> y_true = [0, 1, 2, 3, 4, 5]
    >>> y_pred = [1.1, 2.3, 0.4, 3.9, 3.1, 4.6]
    >>> smape(y_true, y_pred)
    0.7864893577539014

    """
    y_true, y_pred = _convert_to_ndarray(y_true, y_pred)
    _check_input(y_true, y_pred)

    non_normalized_smape = sum(
        np.abs(y_pred - y_true) / (np.abs(y_pred) + np.abs(y_true))
    )
    non_normalized_smape_filled = np.nan_to_num(non_normalized_smape, nan=0)
    smape = (2 / len(y_pred)) * non_normalized_smape_filled
    return smape


def max_error(
    y_true: Union[pd.DataFrame, List, np.ndarray],
    y_pred: Union[pd.DataFrame, List, np.ndarray],
) -> float:
    """Compute the maximum error between two vectors.

    Parameters
    ----------
    y_true : array-like, shape (length, 1), required
        The first vector.

    y_pred : array-like, shape (length, 1), required
        The second vector.

    Returns
    -------
    error : float
        The maximum error between the two vectors.

    Examples
    --------
    >>> from gtime.metrics import max_error
    >>> y_true = [0, 1, 2, 3, 4, 5]
    >>> y_pred = [1.1, 2.3, 0.4, 3.9, 3.1, 4.6]
    >>> max_error(y_true, y_pred)
    1.6

    """
    y_true, y_pred = _convert_to_ndarray(y_true, y_pred)
    _check_input(y_true, y_pred)

    error = np.amax(np.absolute((y_true - y_pred)))
    return error


def mse(
    y_true: Union[pd.DataFrame, List, np.ndarray],
    y_pred: Union[pd.DataFrame, List, np.ndarray],
) -> float:
    """Compute the Mean Squared Error(MSE) between two vectors.

    Parameters
    ----------
    y_true : array-like, shape (length, 1), required.
        The first vector.

    y_pred : array-like, shape (length, 1), required.
        The second vector.

    Returns
    -------
    mse : float
        The Mean Squared Error between the two vectors.
        
    Examples
    --------
    >>> from gtime.metrics import mse
    >>> y_true = [0, 1, 2, 3, 4, 5]
    >>> y_pred = [1.1, 2.3, 0.4, 3.9, 3.1, 4.6]
    >>> mse(y_true, y_pred)
    1.20

    """
    y_true, y_pred = _convert_to_ndarray(y_true, y_pred)
    _check_input(y_true, y_pred)

    sum_squared_error = sum((y_true - y_pred) ** 2)
    mse = sum_squared_error / len(y_true)
    return mse


def rmse(
    y_true: Union[pd.DataFrame, List, np.ndarray],
    y_pred: Union[pd.DataFrame, List, np.ndarray],
) -> float:
    """Compute the Root Mean Squared Error(RMSE) between two vectors.

    Parameters
    ----------
    y_true : array-like, shape (length, 1), required.
        The first vector.

    y_pred : array-like, shape (length, 1), required.
        The second vector.

    Returns
    -------
    rmse : float
        The Root Mean Squared Error between the two vectors.

    Examples
    --------
    >>> from gtime.metrics import rmse
    >>> y_true = [0, 1, 2, 3, 4, 5]
    >>> y_pred = [1.1, 2.3, 0.4, 3.9, 3.1, 4.6]
    >>> rmse(y_true, y_pred)
    1.098

    """

    return np.sqrt(mse(y_true, y_pred))


def log_mse(
    y_true: Union[pd.DataFrame, List, np.ndarray],
    y_pred: Union[pd.DataFrame, List, np.ndarray],
) -> float:
    """Compute the Mean Squared Log Error(MSLE) between two vectors.
    Note: Log_mse accepts only positive numbers as input

    Parameters
    ----------
    y_true : array-like, shape (length, 1), required
        The first vector.

    y_pred : array-like, shape (length, 1), required
        The second vector.
    
    Returns
    -------
    log_mse : float
        The mean squared log error between the two vectors.
        
    Examples
    --------
    >>> from gtime.metrics import log_mse
    >>> y_true = [0, 1, 2, 3, 4, 5]
    >>> y_pred = [1.1, 2.3, 0.4, 3.9, 3.1, 4.6]
    >>> log_mse(y_true, y_pred)
    0.244

    """
    y_true, y_pred = _convert_to_ndarray(y_true, y_pred)
    _check_input(y_true, y_pred)

    if (np.any(y_true < 0)) or (np.any(y_pred < 0)):
        raise ValueError("MSLE can not be used when inputs contain Negative values")
    log_y_true = np.log(y_true + 1)
    log_y_pred = np.log(y_pred + 1)
    log_mse = mse(log_y_true, log_y_pred)
    return log_mse


def rmsle(
    y_true: Union[pd.DataFrame, List, np.ndarray],
    y_pred: Union[pd.DataFrame, List, np.ndarray],
) -> float:
    """Compute the Root Mean Squared Log Error(RMSLE) between two vectors.
    Note: RMSLE accepts only positive numbers as input

    Parameters
    ----------
    y_true : array-like, shape (length, 1), required.
        The first vector.

    y_pred : array-like, shape (length, 1), required.
        The second vector.
    
    Returns
    -------
    rmsle_value : float
        The root mean squared log error between the two vectors.
        
    Examples
    --------
    >>> from gtime.metrics import rmsle
    >>> y_true = [0, 1, 2, 3, 4, 5]
    >>> y_pred = [1.1, 2.3, 0.4, 3.9, 3.1, 4.6]
    >>> rmsle(y_true, y_pred)
    0.49

    """

    return np.sqrt(log_mse(y_true, y_pred))


def r_square(
    y_true: Union[pd.DataFrame, List, np.ndarray],
    y_pred: Union[pd.DataFrame, List, np.ndarray],
) -> float:
    """Compute the R squared (Coefficient of Determination) between two vectors.

    Parameters
    ----------
    y_true : array-like, shape (length, 1), required
        The first vector.

    y_pred : array-like, shape (length, 1), required
        The second vector.

    Returns
    -------
    r_square : float
        The R squared between the two vectors.

    Examples
    --------
    >>> from gtime.metrics import r_square
    >>> y_true = [0, 1, 2, 3, 4, 5]
    >>> y_pred = [1.1, 2.3, 0.4, 3.9, 3.1, 4.6]
    >>> r_square(y_true, y_pred)
    0.586

    """
    y_true, y_pred = _convert_to_ndarray(y_true, y_pred)
    _check_input(y_true, y_pred)

    ss_res = sum((y_true - y_pred) ** 2)
    ss_tot = sum((y_true - np.mean(y_true)) ** 2)
    if not np.any(ss_tot):
        if not np.any(ss_res):
            return 1.0
        else:
            return 0.0
    if np.isnan(ss_res / ss_tot):
        return np.NINF
    r_square = 1 - (ss_res / ss_tot)
    return r_square


def mae(
    y_true: Union[pd.DataFrame, List, np.ndarray],
    y_pred: Union[pd.DataFrame, List, np.ndarray],
) -> float:
    """Compute the Mean Absolute Error(also called, Mean Absolute Deviation(MAD) or Mean Ratio) 
    between two vectors.

    Parameters
    ----------
    y_true : array-like, shape (length, 1), required
        The first vector.

    y_pred : array-like, shape (length, 1), required
        The second vector.

    Returns
    -------
    mae_value : float
        The mean absolute error between the two vectors.

    Examples
    --------
    >>> from gtime.metrics import mae
    >>> y_true = [0, 1, 2, 3, 4, 5]
    >>> y_pred = [1.1, 2.3, 0.4, 3.9, 3.1, 4.6]
    >>> mae(y_true, y_pred)
    1.033

    """
    y_true, y_pred = _convert_to_ndarray(y_true, y_pred)
    _check_input(y_true, y_pred)

    mae_value = np.mean(np.abs(y_pred - y_true))
    return mae_value


def mape(
    y_true: Union[pd.DataFrame, List, np.ndarray],
    y_pred: Union[pd.DataFrame, List, np.ndarray],
) -> float:
    """Compute the Mean Absolute Percentage Error(MAPE) between two vectors.

    Parameters
    ----------
    y_true : array-like, shape (length, 1), required.
        The first vector.

    y_pred : array-like, shape (length, 1), required.
        The second vector.

    Returns
    -------
    mape_value : float
        The mean absolute percentage error between the two vectors.

    Examples
    --------
    >>> from gtime.metrics import mape
    >>> y_true = [1, 1, 2, 3, 4, 5]
    >>> y_pred = [1, 2.3, 0.4, 3.9, 3.1, 4.6]
    >>> mape(y_true, y_pred)
    45.08

    """
    y_true, y_pred = _convert_to_ndarray(y_true, y_pred)
    _check_input(y_true, y_pred)

    mape_value = np.mean(np.abs((y_pred - y_true) / y_true))
    if np.isnan(mape_value):
        raise ValueError("MAPE can not be calculated due to Zero/Zero")
    return mape_value * 100


def gmae(
    y_true: Union[pd.DataFrame, List, np.ndarray],
    y_pred: Union[pd.DataFrame, List, np.ndarray],
) -> float:
    """Compute the Geometric Mean Absolute Error between two vectors.

    Parameters
    ----------
    y_true : array-like, shape (length, 1), required
        The first vector.

    y_pred : array-like, shape (length, 1), required
        The second vector.

    Returns
    -------
    gmae_value : float
        The geometric mean absolute error between the two vectors.

    Examples
    --------
    >>> from gtime.metrics import gmae
    >>> y_true = [0, 1, 2, 3, 6, 5]
    >>> y_pred = [-1, 4, 5, 10, 4, 1]
    >>> gmae(y_true, y_pred)
    2.82

    """
    y_true, y_pred = _convert_to_ndarray(y_true, y_pred)
    _check_input(y_true, y_pred)

    absolute_difference = np.abs(y_pred - y_true)
    return 0 if 0 in absolute_difference else gmean(absolute_difference)



import numpy as np
import pytest
import pandas as pd
if pd.__version__ >= '1.0.0':
    import pandas._testing as testing
else:
    import pandas.util.testing as testing
from gtime.causality import GrangerCausality


# Expected values from results of statstools
@pytest.mark.parametrize(
    "test_input, expected",
    [
        (["ssr_f"], 0.8420421667509344),
        (["ssr_chi2"], 0.8327660223526767),
        (["likelihood_chi2"], 0.8341270186135072),
        (["zero_f"], 0.8420421667508992),
    ],
)
def test_granger_pvalues_ssr_f(test_input, expected):
    # Set random seed, otherwise testing creates a new dataframe each time.
    np.random.seed(12)

    data = testing.makeTimeDataFrame(freq="s", nper=1000)
    granger = (
        GrangerCausality(target_col="B", x_col="A", max_shift=10, statistics=test_input)
        .fit(data)
        .results_[0]
    )

    p_value = granger.values[1]

    # Not exactly equal but up test to 7 digits
    np.testing.assert_almost_equal(p_value, expected, decimal=7)



from random import randint

import numpy as np

from gtime.causality import ShiftedPearsonCorrelation
from gtime.causality.tests.common import make_df_from_expected_shifts


def test_pearson_correlation():
    expected_shifts = [randint(2, 9) * 2 for _ in range(3)]
    df = make_df_from_expected_shifts(expected_shifts)

    spc = ShiftedPearsonCorrelation(target_col="A", max_shift=20)
    spc.fit(df)

    shifts = spc.best_shifts_["A"][1:].values
    np.testing.assert_array_equal(shifts, expected_shifts)


def test_pearson_bootstrap_p_values():
    expected_shifts = [randint(2, 9) * 2 for _ in range(3)]
    df = make_df_from_expected_shifts(expected_shifts)
    shifted_test = ShiftedPearsonCorrelation(
        target_col="A", max_shift=5, bootstrap_iterations=500,
    )
    shifted_test.fit(df)

    pearson_p_values = shifted_test.bootstrap_p_values_
    for col_index in range(len(pearson_p_values.columns)):
        assert pearson_p_values.iloc[col_index, col_index] == 0


def test_pearson_permutation_p_values():
    expected_shifts = [randint(2, 9) * 2 for _ in range(3)]
    df = make_df_from_expected_shifts(expected_shifts)
    shifted_test = ShiftedPearsonCorrelation(
        target_col="A", max_shift=5, permutation_iterations=50,
    )
    shifted_test.fit(df)

    pearson_p_values = shifted_test.permutation_p_values_
    for col_index in range(len(pearson_p_values.columns)):
        assert pearson_p_values.iloc[col_index, col_index] == 0



import numpy as np
import pandas as pd
if pd.__version__ >= '1.0.0':
    import pandas._testing as testing
else:
    import pandas.util.testing as testing
import pytest
from hypothesis import given, strategies as st

from gtime.feature_extraction import (
    Shift,
    MovingAverage,
    Exogenous,
    Polynomial,
    CustomFeature,
    MovingCustomFunction,
)
from gtime.utils.hypothesis.time_indexes import giotto_time_series

df = pd.DataFrame.from_dict({"x": [0, 1, 2, 3, 4, 5]})

shift_class_name = Shift().__class__.__name__
df_shift_1 = pd.DataFrame.from_dict({f"x__{shift_class_name}": [np.nan, 0, 1, 2, 3, 4]})
df_shift_m2 = pd.DataFrame.from_dict(
    {f"x__{shift_class_name}": [2, 3, 4, 5, np.nan, np.nan]}
)
df_shift_0 = pd.DataFrame.from_dict({f"x__{shift_class_name}": [0, 1, 2, 3, 4, 5]})


# FIXME: shift a + shift b = shift a+b instead
class TestShift:
    def _correct_shift(self, df: pd.DataFrame, shift: int) -> pd.DataFrame:
        return df.shift(shift)

    @pytest.mark.parametrize(
        ("shift", "expected"), [(1, df_shift_1), (-2, df_shift_m2), (0, df_shift_0)]
    )
    def test_shift_transform(self, shift, expected):
        shift = Shift(shift=shift)
        testing.assert_frame_equal(shift.fit_transform(df), expected)

    def test_multi_columns_time_shift_feature(self):
        shift = Shift(shift=-2)
        df_multi = pd.DataFrame({"x0": [0, 1, 2, 3, 4, 5], "x1": [7, 8, 9, 10, 11, 12]})

        expected_df = pd.DataFrame.from_dict(
            {
                f"x0__{shift_class_name}": [2, 3, 4, 5, np.nan, np.nan],
                f"x1__{shift_class_name}": [9, 10, 11, 12, np.nan, np.nan],
            }
        )

        testing.assert_frame_equal(shift.fit_transform(df_multi), expected_df)

    @given(
        giotto_time_series(
            start_date=pd.Timestamp(2000, 1, 1), end_date=pd.Timestamp(2010, 1, 1)
        ),
        st.integers(0, 200),
    )
    def test_random_ts_and_shifts(self, df: pd.DataFrame, shift: int):
        shift_feature = Shift(shift=shift)

        df_shifted = shift_feature.fit_transform(df)
        correct_df_shifted = self._correct_shift(df, shift)

        #  testing.assert_frame_equal(correct_df_shifted, df_shifted)


# FIXME: mean(df_k, df_k+1) = dft_k
class TestMovingAverage:
    def _correct_ma(self, df: pd.DataFrame, window_size: int) -> pd.DataFrame:
        return (
            df.rolling(window_size)
            .mean()
            .add_suffix("__" + MovingAverage().__class__.__name__)
        )

    def test_invalid_window_size(self):
        window_size = -1
        df = pd.DataFrame.from_dict({"x0": [0, 1, 2, 3, 4, 5]})

        ma_feature = MovingAverage(window_size=window_size)

        with pytest.raises(ValueError):
            ma_feature.fit_transform(df)

    def test_positive_window_size(self):
        window_size = 2
        df = pd.DataFrame.from_dict({"x": [0, 1, 2, 3, 4, 5]})

        ma_feature = MovingAverage(window_size=window_size)
        df_ma = ma_feature.fit_transform(df)
        output_name = "x__" + ma_feature.__class__.__name__
        expected_df_ma = pd.DataFrame.from_dict(
            {output_name: [np.nan, 0.5, 1.5, 2.5, 3.5, 4.5]}
        )

        testing.assert_frame_equal(expected_df_ma, df_ma, check_names=False)

    def test_multi_columns_window_size(self):
        window_size = 2
        df = pd.DataFrame.from_dict(
            {"x0": [0, 1, 2, 3, 4, 5], "x1": [7, 8, 9, 10, 11, 12]}
        )

        ma_feature = MovingAverage(window_size=window_size)
        feature_name = ma_feature.__class__.__name__

        df_ma = ma_feature.fit_transform(df)
        expected_df_ma = pd.DataFrame(
            {
                f"x0__{feature_name}": [np.nan, 0.5, 1.5, 2.5, 3.5, 4.5],
                f"x1__{feature_name}": [np.nan, 7.5, 8.5, 9.5, 10.5, 11.5],
            }
        )

        testing.assert_frame_equal(expected_df_ma, df_ma, check_names=False)

    @given(
        giotto_time_series(
            start_date=pd.Timestamp(2000, 1, 1), end_date=pd.Timestamp(2010, 1, 1)
        ),
        st.integers(0, 100),
    )
    def test_random_ts_and_window_size(self, df: pd.DataFrame, window_size: int):
        ma_feature = MovingAverage(window_size=window_size)
        df_ma = ma_feature.fit_transform(df)
        expected_df_ma = self._correct_ma(df, window_size)

        testing.assert_frame_equal(expected_df_ma, df_ma)


# FIXME: refactor with pytest parametrize
@pytest.mark.skip(reason="TODO: Use pytest parametrize")
class TestExogenous:
    def _correct_exog(self, exog: pd.DataFrame, df: pd.DataFrame) -> pd.DataFrame:
        return exog.reindex(index=df.index)

    def test_correct_exog_none_method(self):
        method = None
        exog = pd.DataFrame.from_dict({"x0": [0, 1, 2, 3]})
        exog.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 2, 1),
            pd.Timestamp(2000, 3, 1),
            pd.Timestamp(2000, 4, 1),
        ]

        df = pd.DataFrame.from_dict({"x1": [10, 11, 12, 13]})
        df.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 1, 2),
            pd.Timestamp(2000, 1, 3),
            pd.Timestamp(2000, 1, 4),
        ]

        exog_feature = Exogenous(exogenous_time_series=exog, method=method)
        feature_name = exog_feature.__class__.__name__

        new_exog_feature = exog_feature.fit_transform(df)
        expected_exog = pd.DataFrame.from_dict(
            {f"x0__{feature_name}": [0, np.nan, np.nan, np.nan]}
        )
        expected_exog.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 1, 2),
            pd.Timestamp(2000, 1, 3),
            pd.Timestamp(2000, 1, 4),
        ]

        testing.assert_frame_equal(expected_exog, new_exog_feature)

    def test_correct_exog_backfill_method(self):
        method = "backfill"

        exog = pd.DataFrame.from_dict({"x0": [0, 1, 2, 3]})
        exog.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 2, 1),
            pd.Timestamp(2000, 3, 1),
            pd.Timestamp(2000, 4, 1),
        ]

        df = pd.DataFrame.from_dict({"x1": [10, 11, 12, 13]})
        df.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 1, 2),
            pd.Timestamp(2000, 1, 3),
            pd.Timestamp(2000, 1, 4),
        ]

        exog_feature = Exogenous(exogenous_time_series=exog, method=method)

        new_exog_feature = exog_feature.fit_transform(df)
        expected_exog = pd.DataFrame.from_dict({"output_name": [0, 1, 1, 1]})
        expected_exog.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 1, 2),
            pd.Timestamp(2000, 1, 3),
            pd.Timestamp(2000, 1, 4),
        ]

        testing.assert_frame_equal(expected_exog, new_exog_feature)

    def test_correct_exog_pad_method(self):
        method = "pad"
        exog = pd.DataFrame.from_dict({"x_1": [0, 1, 2, 3]})
        exog.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 2, 1),
            pd.Timestamp(2000, 3, 1),
            pd.Timestamp(2000, 4, 1),
        ]

        df = pd.DataFrame.from_dict({"x_2": [10, 11, 12, 13]})
        df.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 1, 2),
            pd.Timestamp(2000, 1, 3),
            pd.Timestamp(2000, 1, 4),
        ]

        exog_feature = Exogenous(exogenous_time_series=exog, method=method)

        new_exog_feature = exog_feature.fit_transform(df)
        expected_exog = pd.DataFrame.from_dict({"output_name": [0, 0, 0, 0]})
        expected_exog.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 1, 2),
            pd.Timestamp(2000, 1, 3),
            pd.Timestamp(2000, 1, 4),
        ]

        testing.assert_frame_equal(expected_exog, new_exog_feature)

    def test_correct_nearest_pad_method(self):
        method = "nearest"
        exog = pd.DataFrame.from_dict({"x_1": [0, 1, 2, 3]})
        exog.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 2, 1),
            pd.Timestamp(2000, 3, 1),
            pd.Timestamp(2000, 4, 1),
        ]

        df = pd.DataFrame.from_dict({"x_2": [10, 11, 12, 13]})
        df.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 1, 2),
            pd.Timestamp(2000, 1, 3),
            pd.Timestamp(2000, 1, 29),
        ]

        exog_feature = Exogenous(exogenous_time_series=exog, method=method)

        new_exog_feature = exog_feature.fit_transform(df)
        expected_exog = pd.DataFrame.from_dict({"output_name": [0, 0, 0, 1]})
        expected_exog.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 1, 2),
            pd.Timestamp(2000, 1, 3),
            pd.Timestamp(2000, 1, 29),
        ]

        testing.assert_frame_equal(expected_exog, new_exog_feature)

    def test_correct_multi_columns_exog(self):
        output_name = "exog"
        exog = pd.DataFrame.from_dict({"x_0": [0, 1, 2, 3], "x_1": [5, 6, 7, 8]})
        exog.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 2, 1),
            pd.Timestamp(2000, 3, 1),
            pd.Timestamp(2000, 4, 1),
        ]

        df = pd.DataFrame.from_dict({"x_2": [10, 11, 12, 13]})
        df.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 1, 2),
            pd.Timestamp(2000, 1, 3),
            pd.Timestamp(2000, 1, 29),
        ]

        exog_feature = Exogenous(exogenous_time_series=exog)

        new_exog_feature = exog_feature.fit_transform(df)
        expected_exog = pd.DataFrame.from_dict(
            {
                f"{output_name}_0": [0, np.nan, np.nan, np.nan],
                f"{output_name}_1": [5, np.nan, np.nan, np.nan],
            }
        )
        expected_exog.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 1, 2),
            pd.Timestamp(2000, 1, 3),
            pd.Timestamp(2000, 1, 29),
        ]

        testing.assert_frame_equal(expected_exog, new_exog_feature)


class TestPolynomial:
    def test_correct_pol_features_single_column(self):
        degree = 3
        df = pd.DataFrame.from_dict({"x": [0, 1, 2, 3]})
        df.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 2, 1),
            pd.Timestamp(2000, 3, 1),
            pd.Timestamp(2000, 4, 1),
        ]

        pol_feature = Polynomial(degree=degree)
        feature_name = pol_feature.__class__.__name__

        pol_df = pol_feature.fit_transform(df)

        expected_pol_df = pd.DataFrame.from_dict(
            {
                f"1__{feature_name}": [1.0, 1.0, 1.0, 1.0],
                f"x0__{feature_name}": [0.0, 1.0, 2.0, 3.0],
                f"x0^2__{feature_name}": [0.0, 1.0, 4.0, 9.0],
                f"x0^3__{feature_name}": [0.0, 1.0, 8.0, 27.0],
            }
        )
        expected_pol_df.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 2, 1),
            pd.Timestamp(2000, 3, 1),
            pd.Timestamp(2000, 4, 1),
        ]

        testing.assert_frame_equal(expected_pol_df, pol_df)

    def test_correct_pol_features_multi_columns(self):
        degree = 2
        df = pd.DataFrame.from_dict({"x_1": [0, 2, 4], "x_2": [1, 3, 5]})
        df.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 2, 1),
            pd.Timestamp(2000, 3, 1),
        ]

        pol_feature = Polynomial(degree=degree)
        feature_name = pol_feature.__class__.__name__

        pol_df = pol_feature.fit_transform(df)

        expected_pol_df = pd.DataFrame.from_dict(
            {
                f"1__{feature_name}": [1.0, 1.0, 1],
                f"x0__{feature_name}": [0.0, 2.0, 4],
                f"x1__{feature_name}": [1.0, 3.0, 5],
                f"x0^2__{feature_name}": [0.0, 4.0, 16.0],
                f"x0 x1__{feature_name}": [0.0, 6.0, 20.0],
                f"x1^2__{feature_name}": [1.0, 9.0, 25.0],
            }
        )
        expected_pol_df.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 2, 1),
            pd.Timestamp(2000, 3, 1),
        ]

        testing.assert_frame_equal(expected_pol_df, pol_df)


class TestMovingCustomFunction:
    def test_correct_moving_custom_function(self):
        df = pd.DataFrame.from_dict({"x_1": [0, 7, 2], "x_2": [2, 10, 4]})
        df.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 2, 1),
            pd.Timestamp(2000, 3, 1),
        ]
        custom_feature = MovingCustomFunction(
            custom_feature_function=np.diff, window_size=2
        )
        custom_output = custom_feature.fit_transform(df)

        feature_name = custom_feature.__class__.__name__
        expected_custom_df = pd.DataFrame.from_dict(
            {
                f"x_1__{feature_name}": [np.nan, 7.0, -5],
                f"x_2__{feature_name}": [np.nan, 8.0, -6],
            }
        )
        expected_custom_df.index = [
            pd.Timestamp(2000, 1, 1),
            pd.Timestamp(2000, 2, 1),
            pd.Timestamp(2000, 3, 1),
        ]

        testing.assert_frame_equal(expected_custom_df, custom_output)


def test_custom_function():
    f = lambda x: x + 1
    df_apply = df.apply(f).rename(columns={"x": "x__CustomFeature"})
    cf = CustomFeature(f)
    testing.assert_equal(df_apply, cf.fit_transform(df))



"""
The :mod:`gtime.causality` module deals with the causality tests for time
series data.
"""

from .linear_coefficient import ShiftedLinearCoefficient
from .pearson_correlation import ShiftedPearsonCorrelation
from .granger_causality import GrangerCausality


__all__ = ["ShiftedLinearCoefficient", "ShiftedPearsonCorrelation", "GrangerCausality"]



from typing import List

import pandas as pd
if pd.__version__ >= '1.0.0':
    import pandas._testing as testing
else:
    import pandas.util.testing as testing


def make_df_from_expected_shifts(expected_shifts: List[int]) -> pd.DataFrame:
    testing.N, testing.K = 500, 1

    df = testing.makeTimeDataFrame(freq="D")
    for sh, k in zip(expected_shifts, range(3)):
        df[f"shift_{k}"] = df["A"].shift(-sh)
    df = df.dropna()

    return df


def shift_df_from_expected_shifts(
    df: pd.DataFrame, expected_shifts: List[int]
) -> pd.DataFrame:
    for sh, k in zip(expected_shifts, range(3)):
        df[f"shift_{k}"] = df[f"shift_{k}"].shift(-sh)
    return df.dropna()



from datetime import datetime
from typing import Union, List, Tuple

import numpy as np
import pandas as pd
import pytest
from hypothesis import given, settings, HealthCheck
from hypothesis.extra.numpy import arrays
from hypothesis.extra.pandas import series as pd_series
from hypothesis.strategies import lists, datetimes, floats, integers
from pandas.testing import assert_series_equal

from gtime.utils.hypothesis.time_indexes import (
    pair_of_ordered_dates,
    series_with_datetime_index,
    series_with_period_index,
    available_freqs,
    series_with_timedelta_index,
)
from gtime.utils.testing_constants import DEFAULT_START, DEFAULT_END
from .utils import (
    compare_output_of_input_sequence_to_expected_one,
    compare_output_of_input_series_to_expected_one,
    transform_series_into_time_index_series,
    transform_time_index_series_into_period_index_series,
    timedelta_index_series_to_period_index_series,
    datetime_index_series_to_period_index_series,
)

PandasDate = Union[datetime, pd.Timestamp, str]


class TestListToTimeIndexSeries:
    @given(lists(floats()), datetimes(DEFAULT_START, DEFAULT_END), available_freqs())
    def test_list_start_freq_as_input(
        self, list_: List[float], start: pd.Timestamp, freq: pd.Timedelta,
    ):
        compare_output_of_input_sequence_to_expected_one(
            list_, start=start, end=None, freq=freq
        )

    @given(lists(floats()), datetimes(DEFAULT_START, DEFAULT_END), available_freqs())
    def test_list_end_freq_as_input(
        self, list_: List[float], end: pd.Timestamp, freq: pd.Timedelta,
    ):
        compare_output_of_input_sequence_to_expected_one(
            list_, start=None, end=end, freq=freq
        )

    @settings(suppress_health_check=(HealthCheck.too_slow,))
    @given(lists(floats()), pair_of_ordered_dates())
    def test_error_with_start_end_as_input(
        self, list_: List[float], start_end: Tuple[pd.Timestamp, pd.Timestamp],
    ):
        with pytest.raises(ValueError):
            compare_output_of_input_sequence_to_expected_one(
                list_, *start_end, freq=None,
            )

    @given(lists(floats()), pair_of_ordered_dates(), available_freqs())
    def test_too_many_parameters(
        self,
        list_: List[float],
        start_end: Tuple[pd.Timestamp, pd.Timestamp],
        freq: pd.Timedelta,
    ):
        with pytest.raises(ValueError):
            compare_output_of_input_sequence_to_expected_one(
                list_, *start_end, freq=freq,
            )

    @given(lists(floats()), datetimes(DEFAULT_START, DEFAULT_END))
    def test_list_and_only_start(
        self, list_: List[float], start: pd.Timestamp,
    ):
        compare_output_of_input_sequence_to_expected_one(
            list_, start=start, end=None, freq=None
        )

    @given(lists(floats()), datetimes(DEFAULT_START, DEFAULT_END))
    def test_list_and_only_end(
        self, list_: List[float], end: pd.Timestamp,
    ):
        compare_output_of_input_sequence_to_expected_one(
            list_, start=None, end=end, freq=None
        )

    @given(lists(floats()), available_freqs())
    def test_list_and_only_freq(
        self, list_: List[float], freq: pd.Timedelta,
    ):
        compare_output_of_input_sequence_to_expected_one(
            list_, start=None, end=None, freq=freq
        )

    @given(lists(floats()))
    def test_only_list(
        self, list_: List[float],
    ):
        compare_output_of_input_sequence_to_expected_one(
            list_, start=None, end=None, freq=None
        )


class TestArrayToTimeIndexSeries:
    @given(
        arrays(np.float64, integers(0, 1000)),
        datetimes(DEFAULT_START, DEFAULT_END),
        available_freqs(),
    )
    def test_array_start_freq_as_input(
        self, array: np.ndarray, start: pd.Timestamp, freq: str,
    ):
        compare_output_of_input_sequence_to_expected_one(
            array, start=start, end=None, freq=freq
        )

    @given(
        arrays(np.float64, integers(0, 1000)),
        datetimes(DEFAULT_START, DEFAULT_END),
        available_freqs(),
    )
    def test_array_end_freq_as_input(
        self, array: np.ndarray, end: pd.Timestamp, freq: str,
    ):
        compare_output_of_input_sequence_to_expected_one(
            array, start=None, end=end, freq=freq
        )

    @given(arrays(np.float64, integers(0, 1000)), pair_of_ordered_dates())
    def test_error_with_start_end_as_input(
        self, array: np.ndarray, start_end: Tuple[pd.Timestamp, pd.Timestamp],
    ):
        with pytest.raises(ValueError):
            compare_output_of_input_sequence_to_expected_one(
                array, *start_end, freq=None,
            )

    @given(
        arrays(np.float64, integers(0, 1000)),
        pair_of_ordered_dates(),
        available_freqs(),
    )
    def test_too_many_parameters(
        self,
        array: np.ndarray,
        start_end: Tuple[pd.Timestamp, pd.Timestamp],
        freq: str,
    ):
        with pytest.raises(ValueError):
            compare_output_of_input_sequence_to_expected_one(
                array, *start_end, freq=freq,
            )

    @given(arrays(np.float64, integers(0, 1000)), datetimes(DEFAULT_START, DEFAULT_END))
    def test_array_and_only_start(
        self, array: np.ndarray, start: pd.Timestamp,
    ):
        compare_output_of_input_sequence_to_expected_one(
            array, start=start, end=None, freq=None
        )

    @given(arrays(np.float64, integers(0, 1000)), datetimes(DEFAULT_START, DEFAULT_END))
    def test_list_and_only_end(
        self, array: np.ndarray, end: pd.Timestamp,
    ):
        compare_output_of_input_sequence_to_expected_one(
            array, start=None, end=end, freq=None
        )

    @given(arrays(np.float64, integers(0, 1000)), available_freqs())
    def test_list_and_only_freq(
        self, array: np.ndarray, freq: pd.Timedelta,
    ):
        compare_output_of_input_sequence_to_expected_one(
            array, start=None, end=None, freq=freq
        )

    @given(arrays(np.float64, integers(0, 1000)))
    def test_only_list(
        self, array: np.ndarray,
    ):
        compare_output_of_input_sequence_to_expected_one(
            array, start=None, end=None, freq=None
        )


class TestPandasSeriesToTimeIndexSeries:
    @given(
        pd_series(dtype=float), datetimes(DEFAULT_START, DEFAULT_END), available_freqs()
    )
    def test_series_start_freq_as_input(
        self, series: pd.Series, start: pd.Timestamp, freq: str,
    ):
        compare_output_of_input_series_to_expected_one(
            series, start=start, end=None, freq=freq,
        )

    @given(
        pd_series(dtype=float), datetimes(DEFAULT_START, DEFAULT_END), available_freqs()
    )
    def test_series_end_freq_as_input(
        self, series: pd.Series, end: pd.Timestamp, freq: str,
    ):
        compare_output_of_input_series_to_expected_one(
            series, start=None, end=end, freq=freq
        )

    @given(pd_series(dtype=float), pair_of_ordered_dates())
    def test_error_with_start_end_as_input(
        self, series: pd.Series, start_end: Tuple[pd.Timestamp, pd.Timestamp],
    ):
        with pytest.raises(ValueError):
            compare_output_of_input_series_to_expected_one(
                series, *start_end, freq=None,
            )

    @given(pd_series(dtype=float), pair_of_ordered_dates(), available_freqs())
    def test_too_many_parameters(
        self,
        series: pd.Series,
        start_end: Tuple[pd.Timestamp, pd.Timestamp],
        freq: str,
    ):
        with pytest.raises(ValueError):
            compare_output_of_input_series_to_expected_one(
                series, *start_end, freq=freq,
            )

    @given(pd_series(dtype=float), datetimes(DEFAULT_START, DEFAULT_END))
    def test_series_and_only_start(
        self, series: pd.Series, start: pd.Timestamp,
    ):
        compare_output_of_input_series_to_expected_one(
            series, start=start, end=None, freq=None
        )

    @given(pd_series(dtype=float), datetimes(DEFAULT_START, DEFAULT_END))
    def test_series_and_only_end(
        self, series: pd.Series, end: pd.Timestamp,
    ):
        compare_output_of_input_series_to_expected_one(
            series, start=None, end=end, freq=None
        )

    @given(pd_series(dtype=float), available_freqs())
    def test_series_and_only_freq(
        self, series: pd.Series, freq: pd.Timedelta,
    ):
        compare_output_of_input_series_to_expected_one(
            series, start=None, end=None, freq=freq
        )

    @given(pd_series(dtype=float))
    def test_only_series(
        self, series: pd.Series,
    ):
        compare_output_of_input_series_to_expected_one(
            series, start=None, end=None, freq=None
        )

    @given(series_with_timedelta_index())
    def test_timedelta_index_as_input(self, timedelta_index_series: pd.Series):
        computed_series = transform_series_into_time_index_series(
            timedelta_index_series
        )
        expected_series = timedelta_index_series
        assert_series_equal(computed_series, expected_series)

    @given(series_with_datetime_index())
    def test_datetime_index_as_input(self, datetime_index_series: pd.Series):
        computed_series = transform_series_into_time_index_series(datetime_index_series)
        expected_series = datetime_index_series
        assert_series_equal(computed_series, expected_series)

    @given(series_with_period_index())
    def test_period_index_as_input(self, period_index_series: pd.Series):
        computed_series = transform_series_into_time_index_series(period_index_series)
        expected_series = period_index_series
        assert_series_equal(computed_series, expected_series)


class TestTimeIndexSeriesToPeriodIndexSeries:
    @settings(suppress_health_check=(HealthCheck.too_slow,))
    @given(series_with_period_index())
    def test_only_period_index_as_input(self, period_index_series: pd.Series):
        computed_series = transform_time_index_series_into_period_index_series(
            period_index_series
        )
        expected_series = period_index_series
        assert_series_equal(computed_series, expected_series)

    @given(series_with_period_index(), available_freqs())
    def test_period_index_and_freq_as_input(
        self, period_index_series: pd.Series, freq: pd.Timedelta
    ):
        computed_series = transform_time_index_series_into_period_index_series(
            period_index_series, freq
        )
        expected_series = period_index_series
        assert_series_equal(computed_series, expected_series)

    @given(series_with_datetime_index())
    def test_only_datetime_index_as_input(self, datetime_index_series: pd.Series):
        computed_series = transform_time_index_series_into_period_index_series(
            datetime_index_series
        )
        expected_series = datetime_index_series_to_period_index_series(
            datetime_index_series
        )
        assert_series_equal(computed_series, expected_series)

    @given(series_with_datetime_index(), available_freqs())
    def test_datetime_index_and_freq_as_input(
        self, datetime_index_series: pd.Series, freq: pd.Timedelta
    ):
        computed_series = transform_time_index_series_into_period_index_series(
            datetime_index_series, freq=freq
        )
        expected_series = datetime_index_series_to_period_index_series(
            datetime_index_series, freq
        )
        assert_series_equal(computed_series, expected_series)

    @given(series_with_timedelta_index())
    def test_only_timedelta_index_as_input(self, timedelta_index_series: pd.Series):
        computed_series = transform_time_index_series_into_period_index_series(
            timedelta_index_series
        )
        expected_series = timedelta_index_series_to_period_index_series(
            timedelta_index_series
        )
        assert_series_equal(computed_series, expected_series)

    def test_basic_timedelta_index_as_input(self):
        timedelta_index_series = pd.Series(
            index=pd.timedelta_range(start=pd.Timedelta(days=1), freq="10D", periods=3),
            data=[1, 2, 3],
        )
        expected_series = pd.Series(
            index=pd.PeriodIndex(
                ["1970-01-02", "1970-01-12", "1970-01-22"], freq="10D"
            ),
            data=[1, 2, 3],
        )
        computed_series = transform_time_index_series_into_period_index_series(
            timedelta_index_series
        )
        assert_series_equal(computed_series, expected_series)

    @given(series_with_timedelta_index(), available_freqs())
    def test_timedelta_index_and_freq_as_input(
        self, timedelta_index_series: pd.Series, freq: pd.Timedelta
    ):
        computed_series = transform_time_index_series_into_period_index_series(
            timedelta_index_series, freq=freq
        )
        expected_series = timedelta_index_series_to_period_index_series(
            timedelta_index_series, freq=freq
        )
        assert_series_equal(computed_series, expected_series)

    @given(pd_series(dtype=float))
    def test_non_time_index_input(self, series: pd.Series):
        with pytest.raises(ValueError):
            transform_time_index_series_into_period_index_series(series)



import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_array, check_X_y, check_random_state


def l1(a, b):
    return np.abs(np.subtract(a, b))


class HedgeForecaster(BaseEstimator):
    """Regressor model using Hedge algorithm.

    This algorithm is based on a multiplicative weight update method to create a dynamic combination of regressive
    models. In theory, there is no common training phase on data, only the loss is necessary to update the model.

    Parameters
    ----------

    learning_rate : float, (default=0.001)
        The factor to use for the weight update.

    loss : callable, optional (default=`gtime.forecasting.online.l1`)
        Loss function use to compute loss matrix.

    random_state : int, RandomState instance or None, optional (default=None)
        Controls both the randomness of the bootstrapping of the samples used
        when building trees (if ``bootstrap=True``) and the sampling of the
        features to consider when looking for the best split at each node
        (if ``max_features < n_features``).
        # TODO: write glossary
        See :term:`Glossary <random_state>` for details.

    Attributes
    ----------
    loss_matrix_ : array, (n_samples, n_experts)
        Loss matrix between X and y.

    total_loss_ : int or float,
        Sum of losses based on Hedge algorithm decisions.

    weights_ : array, (n_experts)
        Last weight of each expert.

    decisions_ : array, (n_samples)
        Indices of chosen expert depending on weights.

    Examples
    --------
    >>> import pandas as pd
    >>> import numpy as np
    >>> from gtime.forecasting.online import HedgeForecaster
    >>> time_index = pd.date_range("2020-01-01", "2020-01-20")
    >>> X = pd.DataFrame(np.random.randint(4, size=(20, 3)), index=time_index)
    >>> y = pd.DataFrame(np.random.randint(4, size=(20, 1)), index=time_index, columns=["y_1"])
    >>> hr = HedgeForecaster(random_state=42)
    >>> hr.fit_predict(X, y).head()
                0
    2020-01-01  2
    2020-01-02  0
    2020-01-03  3
    2020-01-04  3
    2020-01-05  2
    >>> print(f"Estimator weights: {hr.weights_}")
    Estimator weights: [0.97713925 0.97723619 0.97980439]
    >>> print(f"Decisions: {hr.decisions_}")
    Decisions: [1 2 2 1 0 0 0 2 1 2 0 2 2 0 0 0 0 1 1 0]
    >>> print(f"Total loss: {hr.total_loss_}")
    Total loss: 30

    """

    def __init__(
        self, learning_rate: float = 0.001, loss: callable = l1, random_state=None
    ):
        self.eps = learning_rate
        self.loss = loss
        self.random_state = random_state
        pass

    def hedge(self, timestamps, n_experts, loss, eps, random_state):
        weights = np.ones(n_experts)
        self.decisions_ = np.zeros(timestamps, dtype=int)

        total_loss = 0
        for t in range(timestamps):
            self.decisions_[t] = random_state.choice(
                n_experts, p=weights / np.sum(weights)
            )
            total_loss += loss[t][np.int(self.decisions_[t])]
            weights *= np.exp(-eps * loss[t])
        return total_loss, weights

    def fit(self, X, y):
        """ Fit the model to data, compute weights and decisions iteratively.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Data.

        Returns
        -------
        self : object
        """

        random_state = check_random_state(self.random_state)

        #  FIXME: multi_output is not currently supported but mono-column dataframe is 2D (n, 1) so multi_output=True
        #  makes it easier to handle
        X, y = check_X_y(X, y, multi_output=True, y_numeric=True)

        self.loss_matrix_ = self.loss(X, y)

        timestamps = len(X)
        n_experts = X.shape[1]

        self.total_loss_, self.weights_ = self.hedge(
            timestamps=timestamps,
            n_experts=n_experts,
            loss=self.loss_matrix_,
            eps=self.eps,
            random_state=random_state,
        )

        return self

    def fit_predict(self, X, y):
        """Fit and predict variable using Hedge algorithm.

        Parameters
        ----------
        X : (sparse) array-like, shape (n_samples, n_features)
            Data.

        y : (sparse) array-like, shape (n_samples, n_outputs)
            Predictions.

        Returns
        -------
        predictions : pd.DataFrame
            Predictions.
        """
        self.fit(X, y)

        predictions = pd.DataFrame(
            np.take_along_axis(check_array(X), self.decisions_.reshape(-1, 1), axis=1),
            index=X.index,
        )

        return predictions



from datetime import datetime
from typing import List, Union, Optional

import numpy as np
import pandas as pd

from .time_series_conversion import (
    _PandasSeriesToTimeIndexSeries,
    _SequenceToTimeIndexSeries,
    _TimeIndexSeriesToPeriodIndexSeries,
)
from ..preprocessing.time_series_resampling import _TimeSeriesResampler

SUPPORTED_SEQUENCE_TYPES = [
    np.ndarray,
    list,
]


class TimeSeriesPreparation:
    """Transforms an array-like sequence in a period-index DataFrame with a single
    column.

    Here is what happens:
    - if a `list` or `np.array` is passed, the PeriodIndex is built using the parameters
        `start`, `end` and `freq`
    - if a `pd.Series` is passed, it checks if the index is a time index
       (`DatetimeIndex`, `TimedeltaIndex`, `PeriodIndex`) or not. If not the index is
       built as if it were a `list` or `np.array. If yes the index is converted to
       PeriodIndex.

    Parameters
    ----------
    start : datetime, optional, default: ``None``
        The date to use as start date.

    end : datetime, optional, default: ``None``
        The date to use as end date.

    freq : pd.Timedelta, optional, default: ``None``
        The frequency of the output time series. Not mandatory for all time series
        conversion.

    resample_if_not_equispaced : bool, optional, default: ``False``
        Not supported yet, leave it as True

    output_name : str, optional, default: ``'time_series'``
        The name of the output column

    Raises
    ------
    ValueError
        Of the three parameters: start, end, and periods, exactly two must be specified.

    Examples
    --------
    >>> import pandas as pd
    >>> from gtime.preprocessing import TimeSeriesPreparation
    >>> time_series = [1, 2, 3, 5, 5, 7]
    >>> period_index_time_series = pd.Series(
    ...     index = pd.period_range(start='01-01-2010', freq='10D', periods=6),
    ...     data=[1,2,3,5,5,7]
    ... )
    >>> datetime_index_time_series = pd.Series(
    ...     index = pd.date_range(start='01-01-2010', freq='10D', periods=6),
    ...     data=[1,2,3,5,5,7]
    ... )
    >>> timedelta_index_time_series = pd.Series(
    ...     index = pd.timedelta_range(start=pd.Timedelta(days=1), freq='10D', periods=6),
    ...     data=[1,2,3,5,5,7]
    ... )
    >>> time_series_preparation = TimeSeriesPreparation()
    >>> time_series_preparation.transform(time_series)
                time_series
    1970-01-01            1
    1970-01-02            2
    1970-01-03            3
    1970-01-04            5
    1970-01-05            5
    1970-01-06            7
    >>> time_series_preparation.transform(period_index_time_series)
                time_series
    2010-01-01            1
    2010-01-11            2
    2010-01-21            3
    2010-01-31            5
    2010-02-10            5
    2010-02-20            7
    >>> time_series_preparation.transform(datetime_index_time_series)
                time_series
    2010-01-01            1
    2010-01-11            2
    2010-01-21            3
    2010-01-31            5
    2010-02-10            5
    2010-02-20            7
    >>> time_series_preparation.transform(timedelta_index_time_series)
                time_series
    1970-01-02            1
    1970-01-12            2
    1970-01-22            3
    1970-02-01            5
    1970-02-11            5
    1970-02-21            7

    """

    def __init__(
        self,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        freq: Optional[pd.Timedelta] = None,
        resample_if_not_equispaced: bool = False,
        output_name: str = "time_series",
    ):
        self.start = start
        self.end = end
        self.freq = freq
        self.resample_if_not_equispaced = resample_if_not_equispaced
        self.output_name = output_name

        self.pandas_converter = _PandasSeriesToTimeIndexSeries(
            self.start, self.end, self.freq
        )
        self.sequence_converter = _SequenceToTimeIndexSeries(
            self.start, self.end, self.freq
        )
        self.resampler = _TimeSeriesResampler()
        self.to_period_index_series_converter = _TimeIndexSeriesToPeriodIndexSeries(
            self.freq
        )

    def transform(
        self, time_series: Union[List, np.array, pd.Series, pd.DataFrame]
    ) -> pd.DataFrame:
        """Transforms an array-like sequence in a period-index DataFrame with a single
        column.

        Parameters
        ----------
        time_series : Union[List, np.array, pd.Series, pd.DataFrame], required
            The input time series.

        Returns
        -------
        period_index_dataframe : pd.DataFrame
            The output dataframe with a period index.

        """
        pandas_time_series = self._to_time_index_series(time_series)
        equispaced_time_series = self._to_equispaced_time_series(pandas_time_series)
        period_index_time_series = self._to_period_index_time_series(
            equispaced_time_series
        )
        period_index_dataframe = self._to_period_index_dataframe(
            period_index_time_series
        )
        return period_index_dataframe

    def _to_time_index_series(
        self, array_like_object: Union[List, np.array, pd.Series, pd.DataFrame]
    ) -> pd.Series:
        if isinstance(array_like_object, pd.DataFrame):
            return self.pandas_converter.transform(array_like_object.iloc[:, 0])
        elif isinstance(array_like_object, pd.Series):
            return self.pandas_converter.transform(array_like_object)
        elif any(
            isinstance(array_like_object, type_) for type_ in SUPPORTED_SEQUENCE_TYPES
        ):
            return self.sequence_converter.transform(array_like_object)
        else:
            raise TypeError(
                f"Type {type(array_like_object)} is not a "
                f"supported time series type"
            )

    def _to_equispaced_time_series(self, time_series: pd.Series) -> pd.Series:
        if self.resample_if_not_equispaced:
            self.resampler.transform(time_series)
        else:
            return time_series

    def _to_period_index_time_series(self, time_series: pd.Series) -> pd.Series:
        return self.to_period_index_series_converter.transform(time_series)

    def _to_period_index_dataframe(self, time_series: pd.Series) -> pd.DataFrame:
        return pd.DataFrame({self.output_name: time_series})



from gtime._version import __version__

__all__ = [
    "causality",
    "compose",
    "feature_extraction",
    "feature_generation",
    "forecasting",
    "metrics",
    "model_selection",
    "preprocessing",
    "regressors",
    "utils",
]



import numpy as np
import pandas as pd
from scipy.optimize import minimize
from sklearn.metrics import mean_squared_error
from sklearn.utils.validation import check_is_fitted


class LinearRegressor:
    """Implementation of a LinearRegressor that takes a custom loss function.

    Parameters
    ----------
    loss : Callable, optional, default: ``mean_squared_error``
        The loss function to use when fitting the model. The loss function must accept
        y_true, y_pred and return a single real number.

    Examples
    --------
    >>> from gtime.regressors.linear_regressor import LinearRegressor
    >>> from gtime.metrics import max_error
    >>> import numpy as np
    >>> import pandas as pd
    >>> X = np.random.random((100, 10))
    >>> y = np.random.random(100)
    >>> lr = LinearRegressor(loss=max_error)
    >>> X_train, y_train = X[:90], y[:90]
    >>> X_test, y_test = X[90:], y[90:]
    >>> x0 = [0]*11
    >>> lr.fit(X_train, y_train, x0=x0)
    >>> lr.predict(X_test)
    array([0.62987155, 0.46971378, 0.50421395, 0.5543149 , 0.50848151,
           0.54768797, 0.50968854, 0.50500384, 0.58069366, 0.54912972])

    """

    def __init__(self, loss=mean_squared_error):
        self.loss = loss

    def fit(self, X: pd.DataFrame, y: pd.DataFrame, **kwargs) -> "LinearRegressor":
        """Fit the linear model on ``X`` and ``y`` on the given loss function.To do the
        minimization, the ``scipy.optimize.minimize`` function is used. To have more
        details and check which kind of options are available, please refer to the scipy
        `documentation
        <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_.

        Parameters
        ----------
        X : pd.DataFrame, shape (n_samples, n_features), required
            The X matrix used as features in the fitting procedure.

        y : pd.DataFrame, shape (n_samples, 1), required
            The y matrix to use as target values in the fitting procedure.

        kwargs: dict, optional.
            Optional arguments to pass to the ``minimize`` function of scipy.

        Returns
        -------
        self: LinearRegressor
            The fitted model.

        """

        if isinstance(X, pd.DataFrame):
            X = X.values

        if isinstance(y, pd.DataFrame):
            y = y.values

        def prediction_error(model_weights):
            predictions = [
                model_weights[0] + np.dot(model_weights[1:], row) for row in X
            ]
            return self.loss(y, predictions)

        res = minimize(prediction_error, **kwargs)

        self.model_weights_ = res["x"]

        return self

    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        """Predict the y values associated to the features ``X``.

        Parameters
        ----------
        X : pd.DataFrame, shape (n_samples, n_features), required
            The features used to predict.

        Returns
        -------
        predictions : pd.DataFrame, shape (n_samples, 1)
            The predictions of the model

        """
        check_is_fitted(self)

        predictions = self.model_weights_[0] + np.dot(X, self.model_weights_[1:])
        return predictions



"""
The :mod:`gtime.feature_extraction` module deals with the creation of features
starting from a time series.
"""

from gtime.feature_generation.calendar import Calendar
from .standard import (
    Shift,
    MovingAverage,
    MovingCustomFunction,
    Polynomial,
    Exogenous,
    CustomFeature,
)

from .trend import Detrender

__all__ = [
    "Shift",
    "MovingAverage",
    "MovingCustomFunction",
    "Polynomial",
    "Exogenous",
    "Calendar",
    "Detrender",
    "CustomFeature",
]



import numpy as np
import pandas as pd
import pytest
from hypothesis import given
from pandas.testing import assert_series_equal

from gtime.utils.hypothesis.time_indexes import (
    period_indexes,
    series_with_period_index,
    datetime_indexes,
    series_with_datetime_index,
    timedelta_indexes,
    series_with_timedelta_index,
    available_freqs,
    positive_bounded_integers,
    pair_of_ordered_timedeltas,
    pair_of_ordered_dates,
    samples_from,
)
from gtime.utils.hypothesis.utils import freq_to_timedelta

NON_UNIFORM_FREQS = ["B", "Q", "A"]


class TestPeriodIndex:
    @given(period_indexes())
    def test_period_indexes_is_period(self, index):
        assert isinstance(index, pd.PeriodIndex)

    @given(period_indexes(min_length=10, max_length=1000))
    def test_period_indexes_size(self, index):
        assert 10 <= len(index) <= 1000

    @given(period_indexes(min_length=10, max_length=10))
    def test_period_indexes_size_fixed_value(self, index):
        assert len(index) == 10

    @given(period_indexes(min_length=0, max_length=0))
    def test_period_indexes_size_fixed_value_0(self, index):
        assert len(index) == 0

    @given(period_indexes())
    def test_period_indexes_boundaries(self, index):
        start_datetime = pd.Period("1979-12-31").to_timestamp()
        end_datetime = pd.Period("2020-01-01").to_timestamp()
        if len(index):
            assert index[0].to_timestamp() >= start_datetime
            assert index[-1].to_timestamp() <= end_datetime


class TestPeriodSeries:
    @given(series_with_period_index())
    def test_period_series_has_period_index(self, series):
        assert isinstance(series.index, pd.PeriodIndex)

    @given(series_with_period_index(min_length=10, max_length=1000))
    def test_period_series_size(self, series):
        assert 10 <= len(series.index) <= 1000

    @given(series_with_period_index(min_length=10, max_length=10))
    def test_period_series_size_fixed_value(self, series):
        assert len(series.index) == 10

    @given(series_with_period_index(min_length=0, max_length=0))
    def test_period_series_size_fixed_value_0(self, series):
        assert len(series.index) == 0

    @given(series_with_period_index())
    def test_period_series_boundaries(self, series):
        start_datetime = pd.Period("1979-12-31").to_timestamp()
        end_datetime = pd.Period("2020-01-01").to_timestamp()
        if len(series):
            assert series.index[0].to_timestamp() >= start_datetime
            assert series.index[-1].to_timestamp() <= end_datetime

    @given(series_with_period_index())
    def test_period_series_has_float_values(self, series: pd.Series):
        assert series.dtype == "float64"

    @given(series_with_period_index(allow_nan=False))
    def test_period_series_no_nan(self, series: pd.Series):
        assert_series_equal(series, series.dropna())

    @given(series_with_period_index(allow_infinity=False))
    def test_period_series_no_infinity(self, series: pd.Series):
        assert_series_equal(series, series.replace([np.inf, -np.inf], np.nan))


class TestDatetimeIndex:
    @given(datetime_indexes())
    def test_datetime_indexes_is_datetime(self, index):
        assert isinstance(index, pd.DatetimeIndex)

    @given(datetime_indexes(min_length=30, max_length=1000))
    def test_datetime_indexes_size(self, index):
        assert 30 <= len(index) <= 1000

    @given(datetime_indexes(min_length=30, max_length=30))
    def test_datetime_indexes_size_fixed_value(self, index):
        assert len(index) == 30

    @given(datetime_indexes(min_length=0, max_length=0))
    def test_datetime_indexes_size_fixed_value_0(self, index):
        assert len(index) == 0

    @given(datetime_indexes())
    def test_datetime_indexes_boundaries(self, index):
        start_datetime = pd.Timestamp("1979-12-31")
        end_datetime = pd.Timestamp("2020-01-02")
        if len(index):
            assert index[0] >= start_datetime
            assert index[-1] <= end_datetime


class TestDatetimeSeries:
    @given(series_with_datetime_index())
    def test_datetime_series_has_datetime_index(self, series):
        assert isinstance(series.index, pd.DatetimeIndex)

    @given(series_with_datetime_index(min_length=40, max_length=1000))
    def test_datetime_series_size(self, series):
        assert 40 <= len(series.index) <= 1000

    @given(series_with_datetime_index(min_length=40, max_length=40))
    def test_datetime_series_size_fixed_value(self, series):
        assert len(series.index) == 40

    @given(series_with_datetime_index(min_length=0, max_length=0))
    def test_datetime_series_size_fixed_value_0(self, series):
        assert len(series.index) == 0

    @given(series_with_datetime_index())
    def test_datetime_series_boundaries(self, series):
        start_datetime = pd.Timestamp("1979-12-31")
        end_datetime = pd.Timestamp("2020-01-02")
        if len(series):
            assert series.index[0] >= start_datetime
            assert series.index[-1] <= end_datetime

    @given(series_with_datetime_index())
    def test_datetime_series_has_float_values(self, series: pd.Series):
        assert series.dtype == "float64"

    @given(series_with_datetime_index(allow_nan=False))
    def test_datetime_series_no_nan(self, series: pd.Series):
        assert_series_equal(series, series.dropna())

    @given(series_with_datetime_index(allow_infinity=False))
    def test_datetime_series_no_infinity(self, series: pd.Series):
        assert_series_equal(series, series.replace([np.inf, -np.inf], np.nan))


class TestTimedeltaIndex:
    @given(timedelta_indexes())
    def test_timedelta_indexes_is_timedelta(self, index):
        assert isinstance(index, pd.TimedeltaIndex)

    @given(timedelta_indexes(min_length=30, max_length=1000))
    def test_timedelta_indexes_size(self, index):
        assert 30 <= len(index) <= 1000

    @given(timedelta_indexes(min_length=30, max_length=30))
    def test_timedelta_indexes_size_fixed_value(self, index):
        assert len(index) == 30

    @given(timedelta_indexes(min_length=0, max_length=0))
    def test_timedelta_indexes_size_fixed_value_0(self, index):
        assert len(index) == 0

    @given(timedelta_indexes())
    def test_timedelta_indexes_boundaries(self, index):
        start_timedelta, end_timedelta = pd.Timedelta(0), pd.Timedelta("40Y")
        if len(index):
            assert index[0] >= start_timedelta
            assert index[-1] <= end_timedelta


class TestTimedeltaSeries:
    @given(series_with_timedelta_index())
    def test_timedelta_series_has_timedelta_index(self, series):
        assert isinstance(series.index, pd.TimedeltaIndex)

    @given(series_with_timedelta_index(min_length=15, max_length=1000))
    def test_timedelta_series_size(self, series):
        assert 15 <= len(series.index) <= 1000

    @given(series_with_timedelta_index(min_length=15, max_length=15))
    def test_timedelta_series_size_fixed_value(self, series):
        assert len(series.index) == 15

    @given(series_with_timedelta_index(min_length=0, max_length=0))
    def test_timedelta_series_size_fixed_value_0(self, series):
        assert len(series.index) == 0

    @given(series_with_timedelta_index())
    def test_timedelta_series_boundaries(self, series):
        start_timedelta, end_timedelta = pd.Timedelta(0), pd.Timedelta("41y")
        if len(series):
            assert series.index[0] >= start_timedelta
            assert series.index[-1] <= end_timedelta

    @given(series_with_timedelta_index())
    def test_timedelta_series_has_float_values(self, series: pd.Series):
        assert series.dtype == "float64"

    @given(series_with_timedelta_index(allow_nan=False))
    def test_timedelta_series_no_nan(self, series: pd.Series):
        assert_series_equal(series, series.dropna())

    @given(series_with_timedelta_index(allow_infinity=False))
    def test_timedelta_series_no_infinity(self, series: pd.Series):
        assert_series_equal(series, series.replace([np.inf, -np.inf], np.nan))


class TestGeneric:
    @given(available_freqs())
    def test_available_freqs_is_timedelta(self, frequency):
        assert isinstance(frequency, pd.Timedelta)

    @given(positive_bounded_integers(100000))
    def test_positive_bounded_integers_is_positive(self, integer):
        assert integer >= 0

    @given(pair_of_ordered_timedeltas())
    def test_pair_of_ordered_timedeltas_is_ordered(self, pair):
        assert pair[0] < pair[1]

    @given(pair_of_ordered_dates())
    def test_pair_of_ordered_dates_is_ordered(self, pair):
        assert pair[0] < pair[1]

    @given(samples_from(NON_UNIFORM_FREQS))
    def test_freq_to_timedelta(self, freq: str):
        with pytest.raises(ValueError):
            freq_to_timedelta(freq, approximate_if_non_uniform=False)



import pandas as pd
from sklearn.multioutput import MultiOutputRegressor, RegressorChain


class GAR(MultiOutputRegressor):
    """Generalized Auto Regression model.

    This model is a wrapper of ``sklearn.multioutput.MultiOutputRegressor`` but returns
    a ``pd.DataFrame``.

    Fit one model for each target variable contained in the ``y`` matrix.

    Parameters
    ----------
    estimator : estimator object, required
        The model used to make the predictions step by step. Regressor object such as
        derived from ``RegressorMixin``.

    n_jobs : int, optional, default: ``None``
        The number of jobs to use for the parallelization.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from gtime.forecasting import GAR
    >>> from sklearn.ensemble import RandomForestRegressor
    >>> time_index = pd.date_range("2020-01-01", "2020-01-30")
    >>> X = pd.DataFrame(np.random.random((30, 5)), index=time_index)
    >>> y_columns = ["y_1", "y_2", "y_3"]
    >>> y = pd.DataFrame(np.random.random((30, 3)), index=time_index, columns=y_columns)
    >>> X_train, y_train = X[:20], y[:20]
    >>> X_test, y_test = X[20:], y[20:]
    >>> random_forest = RandomForestRegressor()
    >>> gar = GAR(estimator=random_forest)
    >>> gar.fit(X_train, y_train)
    >>> predictions = gar.predict(X_test)
    >>> predictions.shape
    (10, 3)

    """

    def __init__(self, estimator, n_jobs: int = None):
        super().__init__(estimator, n_jobs)

    def fit(self, X: pd.DataFrame, y: pd.DataFrame, sample_weight=None):
        """Fit the model.

        Train the models, one for each target variable in y.

        Parameters
        ----------
        X : pd.DataFrame, shape (n_samples, n_features), required.
            The data.

        y : pd.DataFrame, shape (n_samples, horizon), required.
            The matrix containing the target variables.

        Returns
        -------
        self : object


        """
        self._y_columns = y.columns
        return super().fit(X, y, sample_weight)

    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        """For each row in ``X``, make a prediction for each fitted model, from 1 to
        ``horizon``.

        Parameters
        ----------
        X : pd.DataFrame, shape (n_samples, n_features), required
            The data.

        Returns
        -------
        y_p_df : pd.DataFrame, shape (n_samples, horizon)
            The predictions, one for each timestep in horizon.

        """
        y_p = super().predict(X)
        y_p_df = pd.DataFrame(data=y_p, columns=self._y_columns, index=X.index)

        return y_p_df


# TODO: See #99
class GARFF(RegressorChain):
    """Generalized Auto Regression model with feedforward training. This model is a
    wrapper of ``sklearn.multioutput.RegressorChain`` but returns  a ``pd.DataFrame``.

    Fit one model for each target variable contained in the ``y`` matrix, also using the
    predictions of the previous model.

    Parameters
    ----------
    estimator : estimator object, required
        The model used to make the predictions step by step. Regressor object such as
        derived from ``RegressorMixin``.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from gtime.forecasting import GARFF
    >>> from sklearn.ensemble import RandomForestRegressor
    >>> time_index = pd.date_range("2020-01-01", "2020-01-30")
    >>> X = pd.DataFrame(np.random.random((30, 5)), index=time_index)
    >>> y_columns = ["y_1", "y_2", "y_3"]
    >>> y = pd.DataFrame(np.random.random((30, 3)), index=time_index, columns=y_columns)
    >>> X_train, y_train = X[:20], y[:20]
    >>> X_test, y_test = X[20:], y[20:]
    >>> random_forest = RandomForestRegressor()
    >>> garff = GARFF(estimator=random_forest)
    >>> garff.fit(X_train, y_train)
    >>> predictions = garff.predict(X_test)
    >>> predictions.shape
    (10, 3)

    Notes
    -----
    ``sklearn.multioutput.RegressorChain`` order, cv and random_state parameters were
    set to None due to target order importance in a time-series forecasting context.

    """

    def __init__(self, estimator):
        super().__init__(
            base_estimator=estimator, order=None, cv=None, random_state=None
        )

    def fit(self, X: pd.DataFrame, y: pd.DataFrame):
        """Fit the models, one for each time step. Each model is trained on the initial
        set of features and on the true values of the previous steps.

        Parameters
        ----------
        X : pd.DataFrame, shape (n_samples, n_features), required
            The data.

        y : pd.DataFrame, shape (n_samples, horizon), required
            The matrix containing the target variables.

        Returns
        -------
        self : object
            The fitted object.

        """
        self._y_columns = y.columns
        return super().fit(X, y)

    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        """For each row in ``X``, make a prediction for each fitted model, from 1 to
        ``horizon``.

        Parameters
        ----------
        X : pd.DataFrame, shape (n_samples, n_features)
            The data.

        Returns
        -------
        y_p_df : pd.DataFrame, shape (n_samples, horizon)
            The predictions, one for each timestep in horizon.

        """
        y_p = super().predict(X)
        y_p_df = pd.DataFrame(data=y_p, columns=self._y_columns, index=X.index)

        return y_p_df



"""
The :mod:`gtime.model_selection` module deals with model selection.
"""

from .horizon_shift import horizon_shift
from .splitters import FeatureSplitter

__all__ = ["FeatureSplitter", "horizon_shift"]



import warnings
from itertools import product

import numpy as np
import pandas as pd
from scipy import stats
from sklearn.utils.validation import check_is_fitted


class CausalityMixin:
    """ Base class for causality tests. """

    def __init__(self, bootstrap_iterations, permutation_iterations):
        self.bootstrap_iterations = bootstrap_iterations
        self.permutation_iterations = permutation_iterations

    def transform(self, data: pd.DataFrame) -> pd.DataFrame:
        """Shifts each input time series by the amount which optimizes correlation with
        the selected 'target_col' column. If no target column is specified, the first
        column of the DataFrame is taken as the target.

        Parameters
        ----------
        data : pd.DataFrame, shape (n_samples, n_time_series), required
            The DataFrame containing the time series on which to perform the
            transformation.

        Returns
        -------
        data_t : pd.DataFrame, shape (n_samples, n_time_series)
            The DataFrame (Pivot table) of the shifts which maximize the correlation
            between each time series. The shift is indicated in rows.

        """
        check_is_fitted(self)
        data_t = data.copy()

        if self.target_col is None:
            self.target_col = data_t.columns[0]
            warnings.warn(
                "The target column was not specified. Therefore, the first "
                f"column {self.target_col } of the DataFrame was taken as "
                "target column. If you want to transform with respect to "
                "another column, please use it as a target column."
            )

        for col in data_t:
            if col != self.target_col:
                data_t[col] = data_t[col].shift(self.best_shifts_[self.target_col][col])
        if self.dropna:
            data_t = data_t.dropna()

        return data_t

    def _initialize_table(self):
        best_shifts = pd.DataFrame(columns=["x", "y", "shift", "max_corr"])
        column_types = {
            "x": np.float64,
            "y": np.float64,
            "shift": np.int64,
            "max_corr": np.int64,
        }

        if self.bootstrap_iterations:
            best_shifts = best_shifts.reindex(
                best_shifts.columns.tolist() + ["p_values"], axis=1
            )
            column_types["p_values"] = np.float64

        best_shifts = best_shifts.astype(column_types)
        return best_shifts

    def _compute_best_shifts(self, data, shift_func):
        best_shifts = self._initialize_table()

        if self.target_col is None:
            columns_to_shift = [(x, y) for x, y in product(data.columns, repeat=2)]

        else:
            columns_to_shift = [(col, self.target_col) for col in data.columns]

        for (x, y) in columns_to_shift:
            res = shift_func(data, x=x, y=y)
            best_shift = res[1]
            max_corr = res[0]
            tables = {
                "x": x,
                "y": y,
                "shift": best_shift,
                "max_corr": max_corr,
            }
            if self.bootstrap_iterations:
                bootstrap_p_value = self._compute_p_values(
                    data=data, x=x, y=y, shift=best_shift, test_type="bootstrap"
                )
                tables["bootstrap_p_values"] = bootstrap_p_value

            if self.permutation_iterations:
                bootstrap_p_value = self._compute_p_values(
                    data=data, x=x, y=y, shift=best_shift, test_type="permutation"
                )
                tables["permutation_p_values"] = bootstrap_p_value

            best_shifts = best_shifts.append(tables, ignore_index=True,)

        return best_shifts

    def _compute_p_values(self, data, x, y, shift, test_type):
        data_t = data.copy()
        data_t[x] = data_t.shift(shift)[x]
        data_t.dropna(axis=0, inplace=True)
        rhos = []
        n_iterations = (
            self.permutation_iterations
            if test_type == "permutation"
            else self.bootstrap_iterations
        )

        for k in range(n_iterations):
            if test_type == "permutation":
                samples = data_t.sample(frac=1)
            else:
                samples = data_t.sample(n=len(data), replace=True)

            rhos.append(stats.pearsonr(samples[x], samples[y])[0])
        rhos = pd.DataFrame(rhos)
        percentile = stats.percentileofscore(rhos, 0) / 100
        p_value = 2 * (percentile if percentile < 0.5 else 1 - percentile)

        return p_value

    def _create_pivot_tables(self, best_shifts):
        pivot_best_shifts = pd.pivot_table(
            best_shifts, index=["x"], columns=["y"], values="shift"
        )
        max_corrs = pd.pivot_table(
            best_shifts, index=["x"], columns=["y"], values="max_corr"
        )

        pivot_tables = {"best_shifts": pivot_best_shifts, "max_corrs": max_corrs}

        if self.bootstrap_iterations:
            bootstrap_p_values = pd.pivot_table(
                best_shifts, index=["x"], columns=["y"], values="bootstrap_p_values"
            )
            pivot_tables["bootstrap_p_values"] = bootstrap_p_values

        if self.permutation_iterations:
            permutation_p_values = pd.pivot_table(
                best_shifts, index=["x"], columns=["y"], values="permutation_p_values"
            )
            pivot_tables["permutation_p_values"] = permutation_p_values

        return pivot_tables



"""
The :mod:`gtime.feature_generation` module deals with the creation of features that do
not depend on the input data, but just on its index.
"""

from .calendar import Calendar
from .external import PeriodicSeasonal, Constant

__all__ = ["PeriodicSeasonal", "Constant", "Calendar"]
